"""Query account index state

.. moduleauthor:: Louis Holbrook <dev@holbrook.no>
.. pgp:: 0826EDA1702D1E87C6E2875121D2E7BB88C2A746 

"""

# standard imports
import sys
import os
import json
import argparse
import logging

# external imports
import chainlib.eth.cli
from chainlib.chain import ChainSpec
from chainlib.eth.connection import EthHTTPConnection
from chainlib.eth.tx import receipt
from chainlib.eth.constant import ZERO_CONTENT
from chainlib.error import JSONRPCException
from chainlib.eth.cli.arg import (
        Arg,
        ArgFlag,
        process_args,
        )
from chainlib.eth.cli.config import (
        Config,
        process_config,
        )
from chainlib.eth.cli.log import process_log
from chainlib.eth.settings import process_settings
from chainlib.settings import ChainSettings
from chainlib.eth.address import to_checksum_address

# local imports
from eth_accounts_index import AccountsIndex
from eth_accounts_index.registry import AccountRegistry

logging.basicConfig(level=logging.WARNING)
logg = logging.getLogger()


def process_config_local(config, arg, args, flags):
    address = config.get('_POSARG')
    if address != None:
        address = to_checksum_address(address)
    config.add(address, '_ADDRESS')
    return config


arg_flags = ArgFlag()
arg = Arg(arg_flags)
flags = arg_flags.STD_READ | arg_flags.EXEC

argparser = chainlib.eth.cli.ArgumentParser()
argparser = process_args(argparser, arg, flags)
argparser.add_argument('address', type=str, help='Address to add to registry')
args = argparser.parse_args()

logg = process_log(args, logg)

config = Config()
config = process_config(config, arg, args, flags, positional_name='address')
config = process_config_local(config, arg, args, flags)
logg.debug('config loaded:\n{}'.format(config))

settings = ChainSettings()
settings = process_settings(settings, config)
logg.debug('settings loaded:\n{}'.format(settings))


def out_element(e, w=sys.stdout):
    w.write(str(e[1]) + '\n')


def element(ifc, conn, contract_address, address, w=sys.stdout):
    o = ifc.have(contract_address, address)
    r =  conn.do(o)
    have = ifc.parse_have(r)
    out_element((0, address), w)


def ls(ifc, conn, contract_address, w=sys.stdout):
    i = 0
    while True:
        o = ifc.entry(contract_address, i)
        try:
            r = conn.do(o)
            account = ifc.parse_account(r)
            out_element((i, account), w)
            i += 1
        except JSONRPCException as e:
            break


def main():
    conn = settings.get('CONN')
    address = config.get('_ADDRESS')
    c = AccountsIndex(
            settings.get('CHAIN_SPEC')
            )
    if address != None:
        element(
                c,
                conn,
                settings.get('EXEC'),
                address,
                w=sys.stdout,
                )
    else:
        ls(
                c,
                conn,
                settings.get('EXEC'),
                w=sys.stdout,
                )


if __name__ == '__main__':
    main()



# standard imports
import os
import json

import pathlib
# external imports
from chainlib.eth.tx import (
    TxFormat,
)
from chainlib.eth.contract import (
    ABIContractEncoder,
    ABIContractType,
)

# local imports
from .interface import AccountsIndex

moddir = os.path.dirname(__file__)
datadir = os.path.join(moddir, 'data')


class AccountRegistry(AccountsIndex):
    __abi = None
    __bytecode = None

    @staticmethod
    def abi():
        if AccountRegistry.__abi is None:
            with open(os.path.join(datadir, 'AccountsIndex.json'), 'r') as f:
                AccountRegistry.__abi = json.load(f)
        return AccountRegistry.__abi

    @staticmethod
    def bytecode():
        if AccountRegistry.__bytecode is None:
            AccountRegistry.__bytecode = pathlib.Path(
                os.path.join(datadir, 'AccountsIndex.bin')
            ).read_text()
        return AccountRegistry.__bytecode

    @staticmethod
    def gas(code=None):
        return 1200000

    def constructor(self, sender_address):
        code = AccountRegistry.bytecode()
        tx = self.template(sender_address, None, use_nonce=True)
        tx = self.set_code(tx, code)
        return self.build(tx)

    def __single_address_method(self, method, contract_address, sender_address, address, tx_format=TxFormat.JSONRPC):
        enc = ABIContractEncoder()
        enc.method(method)
        enc.typ(ABIContractType.ADDRESS)
        enc.address(address)
        data = enc.get()
        tx = self.template(sender_address, contract_address, use_nonce=True)
        tx = self.set_code(tx, data)
        tx = self.finalize(tx, tx_format)
        return tx

    def add_writer(self, contract_address, sender_address, address, tx_format=TxFormat.JSONRPC):
        return self.__single_address_method('addWriter', contract_address, sender_address, address, tx_format)

    def delete_writer(self, contract_address, sender_address, address, tx_format=TxFormat.JSONRPC):
        return self.__single_address_method('deleteWriter', contract_address, sender_address, address, tx_format)



# standard imports
import os
import unittest
import json
import logging

# external imports
from chainlib.eth.unittest.ethtester import EthTesterCase
from chainlib.connection import RPCConnection
from chainlib.eth.nonce import RPCNonceOracle
from chainlib.eth.address import to_checksum_address
from chainlib.eth.tx import (
        receipt,
        transaction,
        TxFormat,
        )
from chainlib.eth.contract import (
        abi_decode_single,
        ABIContractType,
        )
from chainlib.eth.block import (
        block_latest,
        block_by_number,
        )

# local imports
from eth_accounts_index.registry import AccountRegistry
from eth_accounts_index import AccountsIndex
from eth_accounts_index.unittest import TestAccountsIndex

logging.basicConfig(level=logging.DEBUG)
logg = logging.getLogger()

testdir = os.path.dirname(__file__)


class Test(TestAccountsIndex):

    def setUp(self):
        super(Test, self).setUp()


    def test_1_count(self):
        #o = self.o.count(self.address, sender_address=self.accounts[0])
        c = AccountsIndex(self.chain_spec)
        o = c.entry_count(self.address, sender_address=self.accounts[0])
        r = self.conn.do(o)
        r = abi_decode_single(ABIContractType.UINT256, r)
        self.assertEqual(r, 0)


    def test_2_add(self):
        b = os.urandom(20)
        a = to_checksum_address(b.hex())
        b = os.urandom(20)
        b = to_checksum_address(b.hex())

        nonce_oracle = RPCNonceOracle(self.accounts[0], self.conn)
        c = AccountsIndex(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash, o) = c.add(self.address, self.accounts[0], a)
        self.conn.do(o)
        o = receipt(tx_hash)
        r = self.conn.do(o)
        self.assertEqual(r['status'], 1)

        self.helper.mine_block()
        o = c.have(self.address, a, sender_address=self.accounts[0])
        r = self.conn.do(o)
        self.assertEqual(int(r, 16), 1)

        o = c.have(self.address, b, sender_address=self.accounts[0])
        r = self.conn.do(o)
        self.assertEqual(int(r, 16), 0)


    def test_3_add_rlpsigned(self):
        b = os.urandom(20)
        a = to_checksum_address(b.hex())

        nonce_oracle = RPCNonceOracle(self.accounts[0], self.conn)
        c = AccountsIndex(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)
        (tx_hash, o) = c.add(self.address, self.accounts[0], a) #, tx_format=TxFormat.RLP_SIGNED)
        r = self.conn.do(o)
        self.assertEqual(tx_hash, r)
        logg.debug('o {}'.format(o))


    def test_4_indices(self):
        nonce_oracle = RPCNonceOracle(self.accounts[0], self.conn)
        c = AccountsIndex(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)

        o = c.entry_count(self.address, sender_address=self.accounts[0])
        r = self.conn.do(o)
        count = c.parse_entry_count(r)
        self.assertEqual(count, 0)

        b = os.urandom(20)
        a = to_checksum_address(b.hex())
        (tx_hash, o) = c.add(self.address, self.accounts[0], a)
        r = self.conn.do(o)

        b = os.urandom(20)
        aa = to_checksum_address(b.hex())
        (tx_hash, o) = c.add(self.address, self.accounts[0], aa)
        r = self.conn.do(o)

        o = c.entry_count(self.address, sender_address=self.accounts[0])
        r = self.conn.do(o)
        count = c.parse_entry_count(r)
        self.assertEqual(count, 2)


    def test_5_deactivate(self):
        nonce_oracle = RPCNonceOracle(self.accounts[0], self.conn)
        c = AccountsIndex(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)

        b = os.urandom(20)
        a = to_checksum_address(b.hex())
        (tx_hash, o) = c.add(self.address, self.accounts[0], a)
        r = self.conn.do(o)

        o = c.entry_count(self.address, sender_address=self.accounts[0])
        r = self.conn.do(o)
        count = c.parse_entry_count(r)
        self.assertEqual(count, 1)

        b = os.urandom(20)
        a = to_checksum_address(b.hex())
        (tx_hash, o) = c.add(self.address, self.accounts[0], a)
        r = self.conn.do(o)

        o = c.entry_count(self.address, sender_address=self.accounts[0])
        r = self.conn.do(o)
        count = c.parse_entry_count(r)
        self.assertEqual(count, 2)

        (tx_hash, o) = c.deactivate(self.address, self.accounts[0], a)
        r = self.conn.do(o)
        o = receipt(r)
        r = self.conn.do(o)
        self.assertEqual(r['status'], 1)

        o = c.entry_count(self.address, sender_address=self.accounts[0])
        r = self.conn.do(o)
        count = c.parse_entry_count(r)
        self.assertEqual(count, 2)

        o = c.have(self.address, a, sender_address=self.accounts[0])
        r = self.conn.do(o)
        self.assertEqual(int(r, 16), 1)

        o = c.is_active(self.address, a, sender_address=self.accounts[0])
        r = self.conn.do(o)
        self.assertEqual(int(r, 16), 0)

        (tx_hash, o) = c.activate(self.address, self.accounts[0], a)
        r = self.conn.do(o)
        o = receipt(r)
        r = self.conn.do(o)
        self.assertEqual(r['status'], 1)

        o = c.have(self.address, a, sender_address=self.accounts[0])
        r = self.conn.do(o)
        self.assertEqual(int(r, 16), 1)

        o = c.is_active(self.address, a, sender_address=self.accounts[0])
        r = self.conn.do(o)
        self.assertEqual(int(r, 16), 1)


    def test_6_remove(self):
        nonce_oracle = RPCNonceOracle(self.accounts[0], self.conn)
        c = AccountsIndex(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)

        b = os.urandom(20)
        a = to_checksum_address(b.hex())
        (tx_hash, o) = c.add(self.address, self.accounts[0], a)
        r = self.conn.do(o)

        o = c.entry_count(self.address, sender_address=self.accounts[0])
        r = self.conn.do(o)
        self.assertEqual(int(r, 16), 1)

        (tx_hash, o) = c.remove(self.address, self.accounts[0], a)
        r = self.conn.do(o)
        o = receipt(r)
        r = self.conn.do(o)
        self.assertEqual(r['status'], 1)

        o = c.entry_count(self.address, sender_address=self.accounts[0])
        r = self.conn.do(o)
        self.assertEqual(int(r, 16), 0)

        addrs = []
        for i in range(8):
            b = os.urandom(20)
            addrs.append(to_checksum_address(b.hex()))
            (tx_hash, o) = c.add(self.address, self.accounts[0], addrs[i])
            r = self.conn.do(o)
            o = receipt(tx_hash)
            r = self.conn.do(o)
            self.assertEqual(r['status'], 1)

        o = c.entry_count(self.address, sender_address=self.accounts[0])
        r = self.conn.do(o)
        self.assertEqual(int(r, 16), 8)

        (tx_hash, o) = c.remove(self.address, self.accounts[0], addrs[4])
        r = self.conn.do(o)
        o = receipt(r)
        r = self.conn.do(o)
        self.assertEqual(r['status'], 1)

        r = []
        for i in range(7):
            o = c.entry(self.address, i, sender_address=self.accounts[0])
            r.append(self.conn.do(o))

        self.assertEqual(len(r), 7)
        self.assertNotIn(addrs[4], r)


    def test_7_time(self):
        nonce_oracle = RPCNonceOracle(self.accounts[0], self.conn)
        c = AccountsIndex(self.chain_spec, signer=self.signer, nonce_oracle=nonce_oracle)

        b = os.urandom(20)
        a = to_checksum_address(b.hex())
        (tx_hash, o) = c.add(self.address, self.accounts[0], a)
        self.conn.do(o)

        o = block_latest()
        r = self.conn.do(o)
        o = block_by_number(r)
        r = self.conn.do(o)
        t = r['timestamp']

        o = c.time(self.address, a, sender_address=self.accounts[0])
        r = self.conn.do(o)
        self.assertEqual(t, int(r, 16))


if __name__ == '__main__':
    unittest.main()



# standard imports
import logging
import json
import os

# external imports
from chainlib.eth.tx import (
    TxFactory,
    TxFormat,
)
from chainlib.eth.contract import (
    ABIContractEncoder,
    ABIContractDecoder,
    ABIContractType,
    abi_decode_single,
)
from chainlib.eth.constant import ZERO_ADDRESS
from chainlib.jsonrpc import JSONRPCRequest
from chainlib.eth.error import RequestMismatchException
from hexathon import (
    add_0x,
    strip_0x,
)

logg = logging.getLogger()


class AccountsIndex(TxFactory):

    def __single_address_method(self, method, contract_address, sender_address, address, tx_format=TxFormat.JSONRPC):
        enc = ABIContractEncoder()
        enc.method(method)
        enc.typ(ABIContractType.ADDRESS)
        enc.address(address)
        data = enc.get()
        tx = self.template(sender_address, contract_address, use_nonce=True)
        tx = self.set_code(tx, data)
        tx = self.finalize(tx, tx_format)
        return tx

    def add(self, contract_address, sender_address, address, tx_format=TxFormat.JSONRPC):
        return self.__single_address_method('add', contract_address, sender_address, address, tx_format)

    def remove(self, contract_address, sender_address, address, tx_format=TxFormat.JSONRPC):
        return self.__single_address_method('remove', contract_address, sender_address, address, tx_format)

    def activate(self, contract_address, sender_address, address, tx_format=TxFormat.JSONRPC):
        return self.__single_address_method('activate', contract_address, sender_address, address, tx_format)

    def deactivate(self, contract_address, sender_address, address, tx_format=TxFormat.JSONRPC):
        return self.__single_address_method('deactivate', contract_address, sender_address, address, tx_format)


    def time(self, contract_address, address, sender_address=ZERO_ADDRESS, id_generator=None):
        j = JSONRPCRequest(id_generator)
        o = j.template()
        o['method'] = 'eth_call'
        enc = ABIContractEncoder()
        enc.method('time')
        enc.typ(ABIContractType.ADDRESS)
        enc.address(address)
        data = add_0x(enc.get())
        tx = self.template(sender_address, contract_address)
        tx = self.set_code(tx, data)
        o['params'].append(self.normalize(tx))
        return o



    def have(self, contract_address, address, sender_address=ZERO_ADDRESS, id_generator=None):
        j = JSONRPCRequest(id_generator)
        o = j.template()
        o['method'] = 'eth_call'
        enc = ABIContractEncoder()
        enc.method('have')
        enc.typ(ABIContractType.ADDRESS)
        enc.address(address)
        data = add_0x(enc.get())
        tx = self.template(sender_address, contract_address)
        tx = self.set_code(tx, data)
        o['params'].append(self.normalize(tx))
        return o

    def is_active(self, contract_address, address, sender_address=ZERO_ADDRESS, id_generator=None):
        j = JSONRPCRequest(id_generator)
        o = j.template()
        o['method'] = 'eth_call'
        enc = ABIContractEncoder()
        enc.method('isActive')
        enc.typ(ABIContractType.ADDRESS)
        enc.address(address)
        data = add_0x(enc.get())
        tx = self.template(sender_address, contract_address)
        tx = self.set_code(tx, data)
        o['params'].append(self.normalize(tx))
        return o


    def entry_count(self, contract_address, sender_address=ZERO_ADDRESS, id_generator=None):
        j = JSONRPCRequest(id_generator)
        o = j.template()
        o['method'] = 'eth_call'
        enc = ABIContractEncoder()
        enc.method('entryCount')
        data = add_0x(enc.get())
        tx = self.template(sender_address, contract_address)
        tx = self.set_code(tx, data)
        o['params'].append(self.normalize(tx))
        return o

    def count(self, contract_address, sender_address=ZERO_ADDRESS, id_generator=None):
        return self.entry_count(contract_address, sender_address=sender_address, id_generator=id_generator)
    
    def entry(self, contract_address, idx, sender_address=ZERO_ADDRESS, id_generator=None):
        j = JSONRPCRequest(id_generator)
        o = j.template()
        o['method'] = 'eth_call'
        enc = ABIContractEncoder()
        enc.method('entry')
        enc.typ(ABIContractType.UINT256)
        enc.uint256(idx)
        data = add_0x(enc.get())
        tx = self.template(sender_address, contract_address)
        tx = self.set_code(tx, data)
        o['params'].append(self.normalize(tx))
        return o

    @classmethod
    def parse_account(self, v):
        return abi_decode_single(ABIContractType.ADDRESS, v)

    @classmethod
    def parse_entry_count(self, v):
        return abi_decode_single(ABIContractType.UINT256, v)

    @classmethod
    def parse_have(self, v):
        return abi_decode_single(ABIContractType.BOOLEAN, v)

    @classmethod
    def parse_add_request(self, v):
        v = strip_0x(v)
        cursor = 0
        enc = ABIContractEncoder()
        enc.method('add')
        enc.typ(ABIContractType.ADDRESS)
        r = enc.get()
        l = len(r)
        m = v[:l]
        if m != r:
            raise RequestMismatchException(v)
        cursor += l

        dec = ABIContractDecoder()
        dec.typ(ABIContractType.ADDRESS)
        dec.val(v[cursor:cursor + 64])
        r = dec.decode()
        return r
